from typing import Any, Any, Mapping, Protocol, Sequence, cast, TYPE_CHECKING
from abc import ABC

from mypy_boto3_bedrock_runtime.type_defs import (
    ConverseResponseTypeDef,
    GuardrailConfigurationTypeDef,
    InferenceConfigurationTypeDef,
    MessageOutputTypeDef,
    MessageUnionTypeDef,
    SystemContentBlockTypeDef,
    ToolConfigurationTypeDef,
)
from typing_extensions import runtime_checkable

from .models import ModelId
from abc import abstractmethod


@runtime_checkable
class BedrockClient(Protocol):
    """Interface for an AWS Bedrock Client. This allows "fake" objects to be passed in for unit tests."""

    @abstractmethod
    def converse(
        self,
        *,
        modelId: str,
        messages: Sequence[MessageUnionTypeDef],
        system: Sequence[SystemContentBlockTypeDef] = ...,
        inferenceConfig: InferenceConfigurationTypeDef = ...,
        toolConfig: ToolConfigurationTypeDef = ...,
        guardrailConfig: GuardrailConfigurationTypeDef = ...,
        additionalModelRequestFields: Mapping[str, Any] = ...,
        additionalModelResponseFieldPaths: Sequence[str] = ...,
    ) -> ConverseResponseTypeDef: ...


class FakeBedrockModel(ABC):
    id: ModelId

    @abstractmethod
    def generate_response(self, messages: Sequence[MessageUnionTypeDef]) -> MessageOutputTypeDef: ...


def empty_list() -> list[Any]:
    return []


def empty_dict() -> dict[Any, Any]:
    return {}


def empty_inference_conf() -> InferenceConfigurationTypeDef:
    return {}


def empty_tool_conf() -> ToolConfigurationTypeDef:
    return {"tools": []}


def empty_guardrails() -> GuardrailConfigurationTypeDef:
    return {
        "guardrailIdentifier": "",
        "guardrailVersion": "",
    }


class FakeBedrockClient:
    """A stub BedrockClient for testing"""

    models: dict[ModelId, FakeBedrockModel]

    def __init__(self, models: Sequence[FakeBedrockModel]):
        self.models = {}
        for model in models:
            self.models[model.id] = model

    def converse(
        self,
        *,
        modelId: str,
        messages: Sequence[MessageUnionTypeDef],
        system: Sequence[SystemContentBlockTypeDef] = empty_list(),
        inferenceConfig: InferenceConfigurationTypeDef = empty_inference_conf(),
        toolConfig: ToolConfigurationTypeDef = empty_tool_conf(),
        guardrailConfig: GuardrailConfigurationTypeDef = empty_guardrails(),
        additionalModelRequestFields: Mapping[str, Any] = empty_dict(),
        additionalModelResponseFieldPaths: Sequence[str] = empty_list(),
    ) -> ConverseResponseTypeDef:
        model = self.models[cast(ModelId, modelId)]
        message = model.generate_response(messages)
        return fake_converse_response(message)


def fake_converse_response(message: MessageOutputTypeDef) -> ConverseResponseTypeDef:
    """Convenience method for testing"""
    return {
        "output": {"message": message},
        "stopReason": "end_turn",
        "metrics": {"latencyMs": 0},
        "additionalModelResponseFields": {},
        "trace": {},
        "ResponseMetadata": {
            "RequestId": "",
            "HTTPStatusCode": 200,
            "HTTPHeaders": {},
            "RetryAttempts": 0,
        },
        "usage": {
            "inputTokens": 0,
            "outputTokens": 0,
            "totalTokens": 0,
        },
    }


# Hack to force mypy to check that Fakes implement the expected interface (Protocol)
if TYPE_CHECKING:
    _client: BedrockClient = FakeBedrockClient([])



from typing import TYPE_CHECKING, Callable, Generic, Optional, TypeVar

from .state_store import StateStore

T = TypeVar("T")


class LocalStateStore(Generic[T]):
    state: dict[str, T]
    default_state: Callable[[], T]

    def __init__(
        self,
        default_state: Callable[[], T],
        initial_state: Optional[dict[str, T]] = None,
    ):
        self.state = initial_state or {}
        self.default_state = default_state

    def get(self, key: str) -> T:
        return self.state.get(key) or self.default_state()

    def set(self, key: str, state: T) -> None:
        self.state[key] = state


if TYPE_CHECKING:
    _store: StateStore[list[str]] = LocalStateStore(lambda: [])



from typing import Generator, Protocol, TypeVar, Callable

from mypy_boto3_bedrock_runtime.type_defs import MessageUnionTypeDef as Message

from ..bedrock import text, maybe_execute_tools
from ..util import Color, color_text, print_message
from .graph import START, Graph, GraphExecutionResult
from ..state import StateStore


class ChatbotContext(Protocol):
    user_id: str
    tools: list[Callable]


class ChatbotState(Protocol):
    messages: list[Message]


T = TypeVar("T")
U = TypeVar("U")


def run_graph(
    graph: Graph[T, U], context: T, store: StateStore[U], store_key: str
) -> Generator[GraphExecutionResult[U], None, None]:
    """
    Executes a Graph[T, U] via a generator, yielding a GraphExecutionResult and control back to the caller each time a layer is executed. Execution occurs in a breadth-first fashion.
    """
    is_finished = False
    last_nodes_executed = [START]

    while is_finished != True:
        state = store.get(store_key)
        result = graph.execute(context, state, from_node=last_nodes_executed)
        last_nodes_executed = result.nodes_executed
        store.set(store_key, result.state)
        is_finished = result.is_finished
        yield result


V = TypeVar("V", bound=ChatbotContext)
W = TypeVar("W", bound=ChatbotState)


def run_chatbot_on_cli(graph: Graph[V, W], context: V, store: StateStore[W]) -> GraphExecutionResult[W]:
    """
    Runs an interactive 'chatbot' on the command line. 'Chatbot' here is defined as a Graph with context (V) and state (W) that conform to the ChatbotContext and ChatbotState Protocols respectively.
    """
    printed_message_index = 0
    last_result: GraphExecutionResult[W]

    for result in run_graph(graph, context, store, context.user_id):
        while printed_message_index < len(result.state.messages):
            message = result.state.messages[printed_message_index]
            if message["role"] == "assistant":
                print_message(message)

            printed_message_index += 1

        last_message = result.state.messages[-1]
        if last_message["role"] == "assistant":
            tool_results = maybe_execute_tools(last_message, context.tools)
            if tool_results:
                result.state.messages = result.state.messages + [tool_results]
            else:
                user_message = input(f"{color_text('User:', Color.GREEN)} ")
                print("\n")
                result.state.messages = result.state.messages + [text(user_message)]

        last_result = result

    return last_result



from inspect import getdoc
from typing import Any, Callable, Dict, Literal, Union, get_args, get_origin, get_type_hints, Optional
from types import UnionType
from pydantic import BaseModel

from mypy_boto3_bedrock_runtime.type_defs import MessageUnionTypeDef as Message
from mypy_boto3_bedrock_runtime.type_defs import (
    ToolResultBlockTypeDef,
    ToolResultContentBlockTypeDef,
    ToolSpecificationTypeDef,
    ToolTypeDef,
    ToolUseBlockOutputTypeDef,
    ToolUseBlockTypeDef,
)

""" 
Code here  adapted from https://github.com/phidatahq/phidata/blob/0cd1431d3025a7ad458bddd15a51500d35c4273d/phi/utils/json_schema.py#L26
"""

JsonType = Literal["string", "number", "integer", "object", "array", "boolean", "null"]

py_to_json_type: dict[str, JsonType] = {
    "int": "number",
    "float": "number",
    "str": "string",
    "bool": "boolean",
    "NoneType": "null",
    "None": "null",
    "object": "object",
    "list": "array",
}


def is_optional_param(param_json_schema: dict[str, Any]) -> bool:
    param_type = param_json_schema["type"]
    match param_type:
        case list():
            return "null" in param_type
        case str():
            return param_type == "null"
        case _:
            return True


def get_json_schema_for_arg(t: Any) -> dict[str, Any]:
    type_args = get_args(t)
    type_origin = get_origin(t)

    match type_origin:
        case list():
            return {"type": "array", "items": get_json_schema_for_arg(type_args[0])}
        case dict():
            return {"type": "object", "properties": {}}
        case None if py_to_json_type.get(t.__name__):
            return {"type": py_to_json_type.get(t.__name__)}
        case None:
            return get_json_schema_from_type_hints(get_type_hints(t))
        case _ if type_origin == Union or type_origin == UnionType:
            return {
                "type": [
                    py_to_json_type.get(arg.__name__) or get_json_schema_from_type_hints(get_type_hints(arg))
                    for arg in type_args
                ]
            }
        case _:
            return {"type": py_to_json_type.get(t.__name__) or get_json_schema_from_type_hints(get_type_hints(t))}


def get_json_schema_from_type_hints(type_hints: Dict[str, Any]) -> Dict[str, Any]:
    schema: Dict[str, Any] = {"type": "object", "properties": {}, "required": []}
    for k, v in type_hints.items():
        if k == "return":
            continue
        arg_json_schema = get_json_schema_for_arg(v)
        schema["properties"][k] = arg_json_schema
        if not is_optional_param(arg_json_schema):
            schema["required"].append(k)
    return schema


def get_tool_spec(f: Callable) -> ToolSpecificationTypeDef:
    type_hints = get_type_hints(f)
    return {
        "name": f.__name__,
        "description": getdoc(f) or "",
        "inputSchema": {"json": get_json_schema_from_type_hints(type_hints)},
    }


def get_tool_defs(tools: list[Callable]) -> list[ToolTypeDef]:
    return [{"toolSpec": get_tool_spec(tool)} for tool in tools]


def maybe_execute_tools(message: Message, tools: list[Callable]) -> Optional[Message]:
    name_to_tool: dict[str, Callable] = {}
    for tool in tools:
        name_to_tool[tool.__name__] = tool

    tool_results = [
        execute_tool(block["toolUse"], name_to_tool) for block in message["content"] if block.get("toolUse")
    ]

    if len(tool_results) > 0:
        return {"role": "user", "content": [{"toolResult": tool_result} for tool_result in tool_results]}

    return None


def execute_tool(
    tool_use: ToolUseBlockTypeDef | ToolUseBlockOutputTypeDef, name_to_tool: dict[str, Callable]
) -> ToolResultBlockTypeDef:
    try:
        tool = name_to_tool[tool_use["name"]]
        tool_result = tool(**tool_use["input"])
        return {
            "toolUseId": tool_use["toolUseId"],
            "content": [tool_result_content_block(tool_result)],
            "status": "success",
        }
    except Exception as e:
        return {"toolUseId": tool_use["toolUseId"], "content": [{"text": str(e)}], "status": "error"}


def tool_result_content_block(result: Any) -> ToolResultContentBlockTypeDef:
    match result:
        case str():
            return {"text": result}
        case int() | float():
            return {"text": str(result)}
        case list():
            return {"json": {"items": [str(r) for r in result]}}
        case dict():
            return {"json": result}
        case BaseModel():
            return {"json": result.model_dump()}
        case _:
            return {"text": str(result)}



from abc import abstractmethod
from typing import Generic, Protocol, TypeVar

T = TypeVar("T")


class StateStore(Protocol, Generic[T]):
    @abstractmethod
    def get(self, key: str) -> T: ...

    @abstractmethod
    def set(self, key: str, state: T) -> None: ...



from .client import BedrockClient, FakeBedrockClient, FakeBedrockModel, fake_converse_response
from .models import ModelId
from .messages import text
from .tools import get_tool_spec, maybe_execute_tools
from .converse import converse, converse_with_structured_output



from typing import Callable, Optional, Type, TypeVar

from mypy_boto3_bedrock_runtime.type_defs import ConverseResponseTypeDef as ConverseResponse
from mypy_boto3_bedrock_runtime.type_defs import InferenceConfigurationTypeDef as InferenceConfig
from mypy_boto3_bedrock_runtime.type_defs import MessageUnionTypeDef as Message
from pydantic import BaseModel

from .client import BedrockClient
from .models import ModelId
from .tools import get_tool_defs


def converse(
    client: BedrockClient,
    model_id: ModelId,
    prompt: str,
    messages: list[Message],
    config: InferenceConfig = {},
    tools: Optional[list[Callable]] = None,
) -> ConverseResponse:
    if tools:
        return client.converse(
            modelId=model_id.value,
            messages=messages,
            system=[{"text": prompt}],
            inferenceConfig=config,
            toolConfig={"tools": get_tool_defs(tools)},
        )

    return client.converse(
        modelId=model_id.value,
        messages=messages,
        system=[{"text": prompt}],
        inferenceConfig=config or {},
    )


T = TypeVar("T", bound=BaseModel, covariant=True)


def converse_with_structured_output(
    client: BedrockClient,
    model_id: ModelId,
    prompt: str,
    messages: list[Message],
    output_schema: Type[T],
    config: Optional[InferenceConfig] = None,
) -> T:
    tool_name = "json_schema"
    response = client.converse(
        modelId=model_id.value,
        messages=messages,
        system=[{"text": prompt}],
        inferenceConfig=config or {},
        toolConfig={
            "tools": [
                {
                    "toolSpec": {
                        "name": f"{tool_name}",
                        "description": "Represents the JSON schema for the desired output format.",
                        "inputSchema": {"json": output_schema.model_json_schema()},
                    }
                }
            ],
            "toolChoice": {"tool": {"name": f"{tool_name}"}},
        },
    )

    json = response["output"]["message"]["content"][0]["toolUse"]["input"]
    return output_schema.model_validate(json)



from enum import Enum

from mypy_boto3_bedrock_runtime.type_defs import MessageUnionTypeDef as Message


class Color(Enum):
    CYAN = "\033[96m"
    GREEN = "\033[92m"


def print_message(message: Message) -> None:
    blocks = [block["text"] for block in message["content"] if block.get("text")]
    text_blocks = "\n".join(blocks)

    match message["role"]:
        case "assistant":
            print(f"{color_text('Assistant: ', Color.CYAN)} {text_blocks}")
        case "user":
            print(f"{color_text('User: ', Color.GREEN)} {text_blocks}")
    print("\n")


def color_text(text: str, color: Color) -> str:
    return f"{color.value}{text}\033[0m"



from typing import Any

from mypy_boto3_bedrock_runtime.literals import ConversationRoleType
from mypy_boto3_bedrock_runtime.type_defs import MessageUnionTypeDef as Message


def text(text_blocks: str | list[str], role: ConversationRoleType = "user") -> Message:
    match text_blocks:
        case list():
            return {"role": role, "content": [{"text": text} for text in text_blocks]}
        case str():
            return {"role": role, "content": [{"text": text_blocks}]}


def tool_result(id: str, results: dict[str, Any]) -> Message:
    return {
        "role": "user",
        "content": [{"toolResult": {"toolUseId": id, "content": [{"text": str(results)}], "status": "success"}}],
    }



from copy import deepcopy
from dataclasses import dataclass
from typing import Callable, Generic, Optional, TypeVar, cast

ID = str
START = "start"
END = "end"


T = TypeVar("T")
U = TypeVar("U")

Node = Callable[[T, U], Optional[U]]
NodeOrId = ID | Node[T, U]
NodeOrNodeWithId = Node[T, U] | tuple[ID, Node[T, U]]

ConditionalEdgeDestination = Callable[[T, U], Optional[NodeOrId[T, U]]]
EdgeDestination = NodeOrId[T, U] | ConditionalEdgeDestination[T, U]
Middleware = Callable[[ID, T], None]


@dataclass
class GraphExecutionResult(Generic[U]):
    state: U
    nodes_executed: list[ID]
    is_finished: bool


class Graph(Generic[T, U]):
    """An immutable Graph. Graphs are executed in a breadth-first fashion."""

    context: T
    root_node: ID
    nodes: dict[ID, Node[T, U]]
    edges: dict[ID, list[EdgeDestination[T, U]]]
    middleware: list[Middleware[U]]

    def __init__(
        self,
        nodes: Optional[list[NodeOrNodeWithId[T, U]]] = None,
        edges: Optional[list[tuple[NodeOrId[T, U], EdgeDestination[T, U]]]] = None,
        middleware: list[Middleware[U]] = [],
    ):
        self.nodes = {}
        self.edges = {}
        self.middleware = middleware

        if nodes:
            for i, n in enumerate(nodes):
                is_root = True if i == 0 else None
                match n:
                    case node if callable(node):
                        self.add_node(node=node, is_root=is_root)
                    case (id, node):
                        self.add_node(node=node, id=id, is_root=is_root)
        if edges:
            for source, destination in edges:
                self.add_edge(source, destination)

    def add_node(self, node: Node[T, U], id: Optional[ID] = None, is_root: Optional[bool] = None) -> None:
        node_id = id if id else node.__name__
        self.nodes[node_id] = node

        if is_root != None:
            self.root_node = node_id
        elif len(self.nodes) == 1:
            self.root_node = node_id

    def add_edge(self, source: NodeOrId[T, U], destination: EdgeDestination[T, U]) -> None:
        source_id = self._get_node_id(source)
        out_edges = self.edges.setdefault(source_id, [])
        out_edges.append(destination)

    def execute(self, context: T, state: U, from_node: list[ID] = [START]) -> GraphExecutionResult[U]:
        """Executes a single layer in the graph and returns a copy of the updated state."""

        state_copy = deepcopy(state)
        nodes_to_execute = self._get_nodes_to_execute(context, state_copy, from_node)

        if nodes_to_execute == [END]:
            return GraphExecutionResult(
                state=state_copy,
                nodes_executed=nodes_to_execute,
                is_finished=True,
            )

        for node_id in nodes_to_execute:
            node = self.nodes[node_id]
            state_copy = node(context, state_copy) or state_copy

        return GraphExecutionResult(
            state=state_copy,
            nodes_executed=nodes_to_execute,
            is_finished=False,
        )

    def _get_nodes_to_execute(self, context: T, state: U, from_node: list[ID]) -> list[ID]:
        if from_node == [START]:
            return [self.root_node]

        nodes_to_execute: list[ID] = []
        for start_node in from_node:
            for node in self._get_connected_nodes(start_node, context, state):
                nodes_to_execute.append(node)

        return nodes_to_execute if len(nodes_to_execute) > 0 else [END]

    def _get_connected_nodes(self, node: ID, context: T, state: U) -> list[ID]:
        children: list[ID] = []
        for edge in self.edges.setdefault(node, []):
            child_id = self._get_destination_id(context, state, edge)
            if child_id:
                children.append(child_id)

        return children

    def _get_destination_id(self, context: T, state: U, edge_destination: EdgeDestination[T, U]) -> Optional[ID]:
        match edge_destination:
            case str():
                return edge_destination

            case function if callable(function):
                if self.nodes.get(function.__name__):
                    return function.__name__
                else:
                    conditional_edge = cast(ConditionalEdgeDestination[T, U], function)
                    node_or_id = conditional_edge(context, state)
                    if node_or_id:
                        return self._get_node_id(node_or_id)
                    else:
                        return None
            case _:
                raise NotImplementedError()

    def _get_node_id(self, node: NodeOrId[T, U]) -> ID:
        return node.__name__ if callable(node) else node
